# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_ensc.ipynb.

# %% auto 0
__all__ = ['EnsemblClient']

# %% ../nbs/04_ensc.ipynb 4
from urllib.parse import urljoin

# %% ../nbs/04_ensc.ipynb 5
from .cons import ENSEMBL_SERVER, RATE_LIMIT
from .reqs import RateLimitRequest

# %% ../nbs/04_ensc.ipynb 7
class EnsemblClient(object):
    rate = RATE_LIMIT
    rest = ENSEMBL_SERVER

    def request(self, endpoint: str, **params) -> RateLimitRequest:
        return RateLimitRequest(urljoin(self.rest, endpoint), rate=self.rate, **params)()
   
    def lookup(self, symbol: str, species: str = 'homo_sapiens', **kwargs):
        endpoint = f'/lookup/symbol/{species}/{symbol}'
        return self.request(endpoint, **kwargs)
    
    def getids(self, symbol: str, species: str = 'homo_sapiens', external_db: str = 'HGNC', **kwargs):
        endpoint = f'/xrefs/symbol/{species}/{symbol}'
        return self.request(endpoint, external_db = external_db, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['RATE_LIMIT', 'RETRY', 'RETRY_AFTER', 'CONTENT_TYPE', 'ENSEMBL_SERVER']

# %% ../nbs/00_cons.ipynb 4
#| export


# %% ../nbs/00_cons.ipynb 6
RATE_LIMIT = 15
RETRY = 'retry'
RETRY_AFTER = 'Retry-After'
CONTENT_TYPE = 'Content-Type'
ENSEMBL_SERVER = 'http://rest.ensembl.org'



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_reqs.ipynb.

# %% auto 0
__all__ = ['JSONRequest', 'RateLimitRequest']

# %% ../nbs/03_reqs.ipynb 4
import sys, json, time
from urllib.parse import urlparse, urlencode, urljoin, ParseResult
from urllib.request import urlopen, Request
from urllib.error import HTTPError

# %% ../nbs/03_reqs.ipynb 5
from .cons import RATE_LIMIT, RETRY_AFTER, CONTENT_TYPE
from .errs import RateLimitError
from .util import RateLimiter

# %% ../nbs/03_reqs.ipynb 7
class JSONRequest(Request):
    def add_query(self, url, **kwargs) -> str:
        scheme, netloc, path, params, _, fragment = list(urlparse(url))
        return ParseResult(scheme, netloc, path, params, urlencode(kwargs), fragment).geturl()
       
    def request(self):
        res = urlopen(self)
        content = res.read()
        if content: data = json.loads(content)
        return data
    
    def __init__(
        self, url, data = None, headers = {}, origin_req_host = None, 
        unverifiable = False, method = None, **params,
    ):
        headers = headers or {}
        headers.setdefault(CONTENT_TYPE, 'application/json')
        url = self.add_query(url, **params)
        super().__init__(url, data, headers, origin_req_host, unverifiable, method)

    def __call__(self, **params):        
        try:
            return self.request()
        except HTTPError as err:
            RateLimitError(err, self.url).handle(self, **params)

# %% ../nbs/03_reqs.ipynb 9
class RateLimitRequest(JSONRequest, Request):
    # All RateLimitRequests share the same rate limiter
    rlim: RateLimiter = RateLimiter(RATE_LIMIT)

    def __init__(
        self, 
        url, data = None, headers = {}, origin_req_host = None, unverifiable = False, method = None, 
        rate: float = RATE_LIMIT, **params
    ):       
        super().__init__(url, data, headers, origin_req_host, unverifiable, method)
        self.rate = rate
        self.rlim.rate = rate # update rate limiter
        self.request = self.rlim(self.request)

    def __call__(self, **params):        
        try: return self.request()
        except HTTPError as err: RateLimitError(err, self.full_url).handle(self, **params)



__version__ = "0.0.4"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00__init__.ipynb 3
from itertools import chain

from . import (cons, errs, util, reqs, ensc, data)
from .cons import *
from .errs import *
from .util import *
from .reqs import *
from .ensc import *
from .data import *

# %% ../nbs/00__init__.ipynb 4
__all__ = sorted(set(chain(*[
    mod.__all__ for mod in (cons, errs, util, reqs, ensc, data)
])))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_errs.ipynb.

# %% auto 0
__all__ = ['RateLimitError']

# %% ../nbs/01_errs.ipynb 4
import sys, time
from urllib.error import HTTPError
from typing import Callable

# %% ../nbs/01_errs.ipynb 5
from .cons import RATE_LIMIT, RETRY_AFTER

# %% ../nbs/01_errs.ipynb 7
class RateLimitError(HTTPError):
    def __init__(self, err: HTTPError, endpoint: str, retry: float = RATE_LIMIT):
        super().__init__(err.url, err.code, err.reason, err.headers, err.fp)
        self.endpoint = endpoint
        self.retry = retry
    
    def handle(self, callback: Callable, **params):
        match self.code:
            case 429:
                retry = self.headers.get(RETRY_AFTER, self.retry)
                time.sleep(float(retry))
                callback(self.endpoint, self.headers, params)
            case _:
                sys.stderr.write(f'Request failed for {self.endpoint}: Status code: {self.code} Reason: {self.reason}\n')



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ensc',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/ensc',
                'lib_path': 'ensc'},
  'syms': { 'ensc.cons': {},
            'ensc.data': { 'ensc.data.GeneInfo': ('data.html#geneinfo', 'ensc/data.py'),
                           'ensc.data.GeneInfo.__call__': ('data.html#geneinfo.__call__', 'ensc/data.py'),
                           'ensc.data.GeneInfo.lookup': ('data.html#geneinfo.lookup', 'ensc/data.py')},
            'ensc.ensc': { 'ensc.ensc.EnsemblClient': ('ensc.html#ensemblclient', 'ensc/ensc.py'),
                           'ensc.ensc.EnsemblClient.getids': ('ensc.html#ensemblclient.getids', 'ensc/ensc.py'),
                           'ensc.ensc.EnsemblClient.lookup': ('ensc.html#ensemblclient.lookup', 'ensc/ensc.py'),
                           'ensc.ensc.EnsemblClient.request': ('ensc.html#ensemblclient.request', 'ensc/ensc.py')},
            'ensc.errs': { 'ensc.errs.RateLimitError': ('errs.html#ratelimiterror', 'ensc/errs.py'),
                           'ensc.errs.RateLimitError.__init__': ('errs.html#ratelimiterror.__init__', 'ensc/errs.py'),
                           'ensc.errs.RateLimitError.handle': ('errs.html#ratelimiterror.handle', 'ensc/errs.py')},
            'ensc.reqs': { 'ensc.reqs.JSONRequest': ('reqs.html#jsonrequest', 'ensc/reqs.py'),
                           'ensc.reqs.JSONRequest.__call__': ('reqs.html#jsonrequest.__call__', 'ensc/reqs.py'),
                           'ensc.reqs.JSONRequest.__init__': ('reqs.html#jsonrequest.__init__', 'ensc/reqs.py'),
                           'ensc.reqs.JSONRequest.add_query': ('reqs.html#jsonrequest.add_query', 'ensc/reqs.py'),
                           'ensc.reqs.JSONRequest.request': ('reqs.html#jsonrequest.request', 'ensc/reqs.py'),
                           'ensc.reqs.RateLimitRequest': ('reqs.html#ratelimitrequest', 'ensc/reqs.py'),
                           'ensc.reqs.RateLimitRequest.__call__': ('reqs.html#ratelimitrequest.__call__', 'ensc/reqs.py'),
                           'ensc.reqs.RateLimitRequest.__init__': ('reqs.html#ratelimitrequest.__init__', 'ensc/reqs.py')},
            'ensc.util': { 'ensc.util.RateLimiter': ('util.html#ratelimiter', 'ensc/util.py'),
                           'ensc.util.RateLimiter.__call__': ('util.html#ratelimiter.__call__', 'ensc/util.py'),
                           'ensc.util.RateLimiter.__init__': ('util.html#ratelimiter.__init__', 'ensc/util.py'),
                           'ensc.util.RateLimiter.__repr__': ('util.html#ratelimiter.__repr__', 'ensc/util.py'),
                           'ensc.util.Singleton': ('util.html#singleton', 'ensc/util.py'),
                           'ensc.util.Singleton.__call__': ('util.html#singleton.__call__', 'ensc/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_data.ipynb.

# %% auto 0
__all__ = ['GeneInfo']

# %% ../nbs/05_data.ipynb 4
import pandas as pd

# %% ../nbs/05_data.ipynb 5
from .ensc import EnsemblClient

# %% ../nbs/05_data.ipynb 7
class GeneInfo:
    ensc = EnsemblClient()
    
    @classmethod
    def lookup(cls, *genes: str, reterrs: bool = False, species: str = 'human', **kwargs):
        info = []
        errs = []
        for gene in genes:
            data = cls.ensc.lookup(gene, species)
            errs.append(gene) if data is None else info.append(data)
        df = pd.DataFrame(info)
        return (df, errs) if reterrs else df
    
    @classmethod
    def __call__(cls, *genes: str, reterrs: bool = False, species: str = 'human', **kwargs):
        return cls.lookup(*genes, reterrs=reterrs, species=species, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_util.ipynb.

# %% auto 0
__all__ = ['Singleton', 'RateLimiter']

# %% ../nbs/02_util.ipynb 4
import time, functools

# %% ../nbs/02_util.ipynb 5
from .cons import RATE_LIMIT

# %% ../nbs/02_util.ipynb 7
class Singleton(type):
    _exists = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._exists:
            cls._exists[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._exists[cls]

class RateLimiter:
    def __init__(self, rate: float = RATE_LIMIT):
        self.rate = rate
        self.calls = []

    def __call__(self, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            now = time.time()
            self.calls = [call for call in self.calls if now - call < 1]
            
            if len(self.calls) >= self.rate:
                time.sleep(1 - (now - self.calls[0]))
            
            response = f(*args, **kwargs)
            self.calls.append(time.time())
            return response
        return wrapped
    
    def __repr__(self) -> str:
        return f'RateLimiter(rate={self.rate}, calls={len(self.calls)})'
